'
'  dd_xml.bas
'
'  DESCRIPTION
'     Don Dickinson's XML parser
'
'  COMPILER
'     Compiles with either PBCC or PBDLL 32-bit
'     Only tested with PBCC2 and PBDLL6, though it
'     is assumed it will work with PBCC1 and PBDLL5
'     There are built-in debug Stdouts that will
'     only work with PBCC. %DEBUG_PBCC needs to be
'     defined in order to turn them on (it is NOT
'     remmed out by default).
'
'  AUTHOR
'     Don Dickinson
'     ddickinson@usinternet.com
'     dickinson.basicguru.com
'
'  LICENSE and DISCLAIMER
'     Copyright 2000-2002
'     Don Dickinson
'     All rights reserved
'
'     This code is free for all to use without acknowledging the author.
'     Use this code as you see fit. By using or compiling this code or derivative
'     thereof, you are consenting to the hold the author, Don Dickinson, harmless
'     for all effects or side-effects its use. This code works great for me,
'     but you are using it at your own risk. You may compile it in your programs
'     and distribute those applications without crediting the author. You may
'     distribute the source code with your application's source code as long
'     as this header remains in place. You may modify it in anyway you find useful.
'     This code (or works derived from it) may be published as long as the header
'     for this file is published and the original author is credited.
'
'  DEPENDS ON
'     pb_mem.bas     Memory handling from Don Dickinson
'     dd_link.bas    Linked list handling from Don Dickinson
'
'  FUNCTIONS
'     xmlParse(incoming)
'        Pass it a string of XML data and it returns a pointer
'        to the root node of the XML Tree. Returns 0 if there
'        was an error. Use xmlLastError() to retrieve the
'        error message.
'     xmlParseFile
'        Loads a file into memory as a set of xml nodes. Returns
'        a pointer to the root node or 0 if errors occur.
'     xmlFreePointers
'        Pass it the root node returned by xmlParse and it frees
'        the memory associated with the that node and recursively
'        all child and sibling nodes.
'     xmlFree
'        Sams as xmlFreePointers
'     xmlLastError
'        Returns a string representation of the last xmlParse Error
'     xmlNodeToString
'        Pass it a node and it returns the XML string for that node
'        not including the end tag.
'     xmlAttribToString
'        Returns the string representation of the attributes
'        for a node.
'     xmlTreeToString
'        Returns the whole XML tree in string form. This is the
'        logical compilment to xmlParse.
'     xmlClean
'        Cleans up text so that it "safe" for xml (removes <, >, etc)
'        by replacing these characters with &gt;, etc
'     xmlUnClean
'        Replaces the &gt; &lt; &apos; &quot; and &amp with
'        their asciiz characters. This is the compliment to xmlClean
'     xmlAttribCount
'        Returns the number of attributes for a node
'     xmlAttribExists
'        Returns %True if the attrib name exists, %False if not
'     xmlAttribName
'        Returns the name of attribute by number
'     xmlAttribValue
'        Returns the value of the attribute given its name
'     xmlFindNode
'        Returns a pointer to a node that has the specified tag
'        and optionally the specified attribute.
'     xmlSetAttrib
'        Updates an existing attributes value or inserts a new
'        attribute if the specified one doesn't exist.
'     xmlAddChild
'        Adds a node as a child of another node
'     xmlAddSimpleChild
'        Adds a child tag node and a single child data node
'        to that one. Returns a pointer to the tag node.
'     xmlGetTag
'        Returns the tag of an xml node in string form.
'     xmlParent
'        Returns pointer to parent node
'     xmlChild
'        Returns a pointer to the first child
'     xmlNextSibling
'        Returns a pointer to the next sibling
'     xmlAttribToList
'        Returns a pointer to a linked list of attributes.
'        This creates a new linked list so the programmer
'        must free it's memory with llFree when done.
'
'  HISTORY
'
'     6-1-2002    v1.2   DD
'        1. Completely re-vamped the xmlParse function.
'           It is now much faster.
'        2. Added xmlGetChildTagData
'           This returns the data from a child tag. For instance ...
'           if the xml is <root><item1>data</item1></root>
'           and you have a pointer to root, xmlGetChildTagData(pRoot,"item1")
'           will return "data"
'
'     9-1-2001    v1.16  DD
'        1. Added xmlParseFile
'        2. Added xmlChildData/xmlExtractSimpleChidData
'        3. Added xmlFindChild/xmlFindDataChild
'
'     8-30-2001   v1.15  DD
'        1. Added xmlAddChild/xmlAddSimpleChild
'        2. Added xmlClean/xmlUnClean
'
'     8-29-2001   v1.14  DD
'        1. Fixed xmlFindNode
'        2. Added xmlAttribExists
'        2. Added xmlSetAttrib
'
'     8-26-2001   v1.13  DD
'        1. Added support for <? ... ?> xml nodes.
'        2. Fixed so that it correctly recognizes
'           carriage returns and tabs as white-space
'           within tags.
'
'     8-22-2001   v1.12  DD
'        Fixed problem with compiler errors when using byval ptr's
'
'     8-19-2001   v1.11  DD
'        <!-- comments -->
'        now supported
'
'     8-18-2001   v1.10  DD
'        1. <tag/> syntax recognized
'        2. single or double quotes allowed as surrounding
'           characters in attribute values
'
'     8-16-2001   v1.00  DD
'        It works.
'        There are two known issues:
'           1. the <tag/> syntax is not yet recognized
'           2. single quotes are not yet allowed as
'              surrounding characters  of attribute
'              values.
'
'  NOTES
'     I am not an XML guru. I know what I've read from w3c.org. I have not
'     done exhaustive reading so I may have missed some things. If you
'     want to learn more, I'd start at http://www.w3c.org or
'     http://www.w3schools.com
'
'     Here are a few important notes:
'        >  XML is not HTML, it has stricter format rules
'        >  Tags *are* case sensitive
'        >  End-tag are not optional (you can use the abbreviated
'           <tag/> format, though.
'        >  You must correctly nest, your tag. This is *not* allowed:
'              <B><I>Something</B></I>
'           it *must* be nested correctly, like this:
'              <B><I><Something</I></B>"
'        >  White space and carriage returns are not ignored. The parser
'           always includes them.
'        >  Attribute values of tags must be enclosed in either single or
'           double quotes. If you use a single quote, then double quotes
'           are allowed in the value. Vice versa is also true.
'        >  Non-tag data is expected to be free of <, >, and &. This
'           goes for attribute values as well. You should replace these with
'           &gt; &lt; &amp;
'
#if not %def(%DD_XML_BAS)

%DD_XML_BAS = 1

'- Required Files
#include "pb_mem.bas"
#include "dd_link.bas"

'- Equates
#if not %def(%False)
%False = 0
#endif
#if not %def(%True)
%True = -1
#endif

'- Conditional compilation
'%DEBUG_PBCC = 1               'rem out this line to suppress PBCC debug messages

'- Globals and Structures
Global dd_xml_Error as String
Global g_xml_Return as String

Type XML_ATTRIB_TYPE
   surroundChar as String * 1          'the surrounding character
   ptrName as Asciiz Ptr               '0-terminated string that names the attribute
   ptrValue as Asciiz Ptr              '0-terminated string of data for the attribute
   ptrNext as XML_ATTRIB_TYPE PTR      'pointer to the next attribute - 0 if no more
End Type

%XML_NODE_INVALID = -1
%XML_NODE_TAG     = 0
%XML_NODE_DATA    = 1
%XML_NODE_COMMENT = 2
%XML_NODE_XML     = 3      ' not yet implemented
%XML_NODE_CDATA   = 4      ' not yet implemented

%XML_NODE_FIND_ANYWHERE = -1
%XML_ANYWHERE = %XML_NODE_FIND_ANYWHERE

Type XML_NODE_TYPE
   nodeType as Long                    'this is either %XML_NODE_TAG,
                                       '%XML_NODE_DATA, or %XML_NODE_COMMENT
   nodeLevel as Long                   'how deep is this? (0 based)
   selfTerminate as Long               'is this a self-terminated tag?
   ptrTag as Asciiz Ptr                'this is the tag name or data
   ptrComment as Asciiz Ptr            'if %XML_NODE_COMMENT, this is the comment
   ptrAttrib as XML_ATTRIB_TYPE PTR    'a pointer to a linked list of XML_ATTRIB_TYPE
   ptrChild as XML_NODE_TYPE PTR       'pointer to the child
   ptrParent as XML_NODE_TYPE PTR      'pointer to the parent
   ptrNextSibling as XML_NODE_TYPE PTR 'pointer to the next sibling
   ptrPrevSibling as XML_NODE_TYPE PTR 'pointer to the next sibling
   ptrNextAlloc as XML_NODE_TYPE PTR   'pointer to next in physical order
End Type

'- Function prototypes
Declare Sub xmlFreeAttribPointers(attribNode as XML_ATTRIB_TYPE)
Declare Sub xmlFreePointers(rootNode as XML_NODE_TYPE Ptr)
Declare Function xmlParse(incoming as String) as DWord

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlLastError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlLastError Alias "xmlLastError" Export as String

   Function = dd_xml_Error

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzLastError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzLastError Alias "xmlzLastError" Export as Dword

   g_xml_Return = dd_xml_Error + $nul
   Function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlClean
'     Removes special characters for strings so that they will
'     display correctly without messing up xml. This is used
'     when the data coming in was typed by a user.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlClean Alias "xmlClean" _
      ( ByVal incoming as String ) Export as String

   Replace "&" with "&amp;" in incoming
   Replace "<" with "&lt;" in incoming
   Replace ">" with "&gt;" in incoming
   Replace chr$(34) with "&quot;" in incoming
   Replace "'" with "&apos;" in incoming

   Function = incoming

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzClean
'     Removes special characters for strings so that they will
'     display correctly without messing up HTML. This is used
'     when the data coming in was typed by a user.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzClean Alias "xmlzClean" _
      ( zIncoming as Asciiz ) Export as Dword

   g_xml_Return = xmlClean( (zIncoming) ) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlUnClean
'     Replaces special xml character tokens with their actual
'     ascii value. This is the "un-do" for xmlClean
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlUnClean Alias "xmlUnClean" _
      ( ByVal incoming as String ) Export as String

   Replace "&apos;" with "'" in incoming
   Replace "&quot;" with chr$(34) in incoming
   Replace "&gt;" with ">" in incoming
   Replace "&lt;" with "<" in incoming
   Replace "&amp;" with "&" in incoming

   Function = incoming

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzUnClean
'     Replaces special html character tokens with their actual
'     ascii value. This is the "un-do" for xmlClean
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzUnClean Alias "xmlzUnClean" _
      ( incoming as Asciiz) Export as Dword

   g_xml_Return = xmlUnClean( (incoming) ) + $nul
   Function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlNodeType
'  Returns a constant representing the node type
'     %XML_NODE_INVALID = -1
'     %XML_NODE_TAG     = 0
'     %XML_NODE_DATA    = 1
'     %XML_NODE_COMMENT = 2
'     %XML_NODE_XML     = 3      ' not yet implemented
'     %XML_NODE_CDATA   = 4      ' not yet implemented
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlNodeType Alias "xmlNodeType" _
      ( node as XML_NODE_TYPE PTR ) Export as Long

   if node = 0 then
      Function = %XML_NODE_INVALID
   else
      Function = @node.nodeType
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribExists
'  Returns a pointer to the attribute structure if
'  the attribute exists, otherwise it returns 0
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribExists Alias "xmlAttribExists" _
      ( node as XML_NODE_TYPE PTR, ByVal sName as String ) Export as Long

   Dim thisAttrib as XML_ATTRIB_TYPE PTR

   sName = trim$(sName)
   if node = 0 then
      Function = 0
      exit function
   end if

   thisAttrib = @node.ptrAttrib
   Do until thisAttrib = 0
      if @thisAttrib.ptrName <> 0 then
         if trim$(@thisAttrib.@ptrName) = sName then
            exit do
         end if
      end if
      thisAttrib = @thisAttrib.ptrNext
   Loop

   Function = thisAttrib

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzAttribExists
'  Returns a pointer to the attribute structure if
'  the attribute exists, otherwise it returns 0
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzAttribExists Alias "xmlzAttribExists" _
      ( node as XML_NODE_TYPE PTR, zName as Asciiz ) Export as Long

   Function = xmlAttribExists(node, (zName) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribCount
'  Returns the number of attributes associated with a node
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribCount Alias "xmlAttribCount" _
      (node as XML_NODE_TYPE PTR) Export as Long

   Dim iCount as Long
   Dim attr as XML_ATTRIB_TYPE PTR

   iCount = 0
   attr = @node.ptrAttrib
   Do until attr = 0
      incr iCount
      attr = @attr.ptrNext
   Loop

   Function = iCount

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribName
'  Returns the attribute's name given it's number (from 1 to xmlAttribCount)
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribName Alias "xmlAttribName" _
      ( node as XML_NODE_TYPE PTR, ByVal attribNumber as Long ) _
      Export as String

   Dim iCount as Long
   Dim nm as String
   Dim attr as XML_ATTRIB_TYPE PTR

   nm = ""
   if attribNumber > 0 then
      attr = @node.ptrAttrib
      Do until attr = 0
         incr iCount
         if iCount = attribNumber then
            nm = @attr.@ptrName
            exit do
         end if
         attr = @attr.ptrNext
      Loop
   end if

   Function = trim$(nm)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzAttribName
'  Returns the attribute's name given it's number (from 1 to xmlAttribCount)
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzAttribName Alias "xmlzAttribName" _
      ( node as XML_NODE_TYPE PTR, ByVal attribNumber as Long ) _
      Export as Dword

   g_xml_return = xmlAttribName(node, attribNumber) + $nul
   function = strptr(g_xml_return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribValue
'  Returns the attribute's value as a string given it's name.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribValue Alias "xmlAttribValue" _
      ( node as XML_NODE_TYPE PTR, ByVal attribName as String ) _
      Export as String

   Dim sFound as String
   Dim attr as XML_ATTRIB_TYPE PTR

   attribName = trim$(attribName)
   sFound = ""
   attr = @node.ptrAttrib
   Do until attr = 0
      if trim$(@attr.@ptrName) = attribName then
         sFound = @attr.@ptrValue
         exit do
      end if
      attr = @attr.ptrNext
   Loop

   Function = sFound

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzAttribValue
'  Returns the attribute's value as a string given it's name.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzAttribValue Alias "xmlzAttribValue" _
      ( node as XML_NODE_TYPE PTR, attribName as Asciiz ) _
      Export as Dword

   g_xml_Return = xmlAttribValue(node, (attribName) ) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlSetAttrib
'  Adds an attribute to the specified node. Returns %True on
'  success and %False on failure.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlSetAttrib Alias "xmlSetAttrib" _
      ( node as XML_NODE_TYPE PTR, ByVal sAName as String, _
        ByVal sAValue as String ) Export as Long

   Dim newAttrib as XML_ATTRIB_TYPE PTR
   Dim lastAttrib as XML_ATTRIB_TYPE PTR

   '- They must specify a name for the attribute.
   sAName = Trim$(sAName)
   if sAName = "" then
      Function = %False
      exit function
   end if

   '- Is this an existing attribute? If so, just
   '  change its value.
   '
   newAttrib = xmlAttribExists(node, sAName)
   if newAttrib <> 0 then
      if @newAttrib.ptrValue then
         FreeMem @newAttrib.ptrValue
      end if
      if sAValue = "" then
         @newAttrib.ptrValue = 0
      else
         @newAttrib.ptrValue = GetAsciiz(sAValue)
      end if
      Function = %True

   '- This attribute doesn't exist for this node, so
   '  add it in.
   '
   else
      newAttrib = GetMem(len(@newAttrib))
      if newAttrib = 0 then
         Function = %False

      else

         '- Assign new attribute
         @newAttrib.ptrName = GetAsciiz(sAName)
         @newAttrib.ptrValue = GetAsciiz(sAValue)
         @newAttrib.surroundChar = $DQ

         '- Add the new attribute to the list
         if @node.ptrAttrib = 0 then
            @node.ptrAttrib = newAttrib
         else
            lastAttrib = @node.ptrAttrib
            Do until @lastAttrib.ptrNext = 0
               lastAttrib = @lastAttrib.ptrNext
            Loop
            @lastAttrib.ptrNext = newAttrib
         end if
         Function = %True

      end if
   end if


End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzSetAttrib
'  Adds an attribute to the specified node. Returns %True on
'  success and %False on failure.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzSetAttrib Alias "xmlzSetAttrib" _
      ( node as XML_NODE_TYPE PTR, zAName as Asciiz, _
        zAValue as Asciiz ) Export as Long

   Function = xmlSetAttrib(node, (zAName), (zAValue) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribToString
'  Formats a node's attributes into a string.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribToString Alias "xmlAttribToString" _
      ( rootNode as XML_NODE_TYPE PTR ) Export as String

   Dim sAccum as String
   Dim ptrAttrib as XML_ATTRIB_TYPE PTR

   if rootNode = 0 then
      Function = ""
      exit function
   end if


   sAccum = ""
   ptrAttrib = @rootNode.ptrAttrib
   do until ptrAttrib = 0
      if (@ptrAttrib.ptrName <> 0)  then
         sAccum = sAccum + " " + @ptrAttrib.@ptrName + "=" + @ptrAttrib.surroundChar
         if (@ptrAttrib.ptrValue <> 0) then
            sAccum = sAccum + xmlClean(@ptrAttrib.@ptrValue) + @ptrAttrib.surroundChar
         else
            sAccum = sAccum + @ptrAttrib.surroundChar
         end if
      end if

      ptrAttrib = @ptrAttrib.ptrNext
   loop

   if Trim$(sAccum) = "" then
      Function = ""
   else
      Function = sAccum
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzAttribToString
'  Formats a node's attributes into a string.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzAttribToString Alias "xmlzAttribToString" _
      ( rootNode as XML_NODE_TYPE PTR ) Export as Dword

   g_xml_Return = xmlAttribToString(rootNode) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlNodeToString
'  Formats a node (including attributes) into string form.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlNodeToString Alias "xmlNodeToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) _
      Export as String

   Dim sAccum as String
   Dim sSpace as String

   if @rootNode.ptrTag = 0 then
      Function = ""
      exit function
   end if

   if iCRLF then
      if @rootNode.nodeLevel < 1 then
         sSpace = ""
      else
         sSpace = Space$(@rootNode.nodeLevel)
      end if
   end if

   if @rootNode.nodeType = %XML_NODE_TAG then
      if @rootNode.selfTerminate then
         sAccum = "<" + @rootNode.@ptrTag + xmlAttribToString(ByVal rootNode) +" />"
      else
         sAccum = "<" + @rootNode.@ptrTag + xmlAttribToString(ByVal rootNode) + ">"
      end if

   elseif @rootNode.nodeType = %XML_NODE_XML then
      sAccum = "<" + @rootNode.@ptrTag + xmlAttribToString(ByVal rootNode) + "?>"

   elseif @rootNode.nodeType = %XML_NODE_COMMENT then
      if @rootNode.ptrComment then
         sAccum = "<!-- " + xmlClean(@rootNode.@ptrComment) + "-->"
      else
         sAccum = "<!-- -->"
      end if

   else
      sAccum = xmlClean(@rootNode.@ptrTag)

   end if

   if iCRLF then
      Function = sSpace + sAccum + $CRLF
   else
      Function = sAccum
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzNodeToString
'  Formats a node (including attributes) into string form.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzNodeToString Alias "xmlzNodeToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) _
      Export as Dword

   g_xml_Return = xmlNodeToString(rootNode, iCRLF) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlTreeToString
'  Formats an entire tree of XML data into its original string form.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlTreeToString Alias "xmlTreeToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) Export as String

   Dim sAccum as String
   Dim sSpace as String

   Do until rootNode = 0
      'if @rootNode.ptrTag = 0 then exit do

      '- Process this node
      sAccum = sAccum + xmlNodeToString(ByVal rootNode, iCRLF)

      '- Process the children of this node recursively
      sAccum = sAccum + xmlTreeToString(ByVal @rootNode.ptrChild, iCRLF)
      if @rootNode.nodeType = %XML_NODE_TAG then
         if @rootNode.selfTerminate = %False then
            if iCRLF then
               if @rootNode.nodeLevel < 1 then
                  sSpace = ""
               else
                  sSpace = Space$(@rootNode.nodeLevel)
               end if
               sAccum = sAccum + sSpace + "</" + @rootNode.@ptrTag + ">" + $CRLF
            else
               sAccum = sAccum + "</" + @rootNode.@ptrTag + ">"
            end if
         end if
      end if

      '- Process the next sibling
      rootNode = @rootNode.ptrNextSibling

   Loop

   Replace $cr with "" in sAccum
   Replace $lf with $crlf in sAccum
   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlBranchToString
'  like xmlTreeToString, but doesn't process siblings
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlBranchToString Alias "xmlBranchToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) Export as String

   Dim sAccum as String
   Dim sSpace as String

   '- Process this node
   sAccum = sAccum + xmlNodeToString(ByVal rootNode, iCRLF)

   '- Process the children of this node recursively
   sAccum = sAccum + xmlTreeToString(ByVal @rootNode.ptrChild, iCRLF)
   if @rootNode.nodeType = %XML_NODE_TAG then
      if @rootNode.selfTerminate = %False then
         if iCRLF then
            if @rootNode.nodeLevel < 1 then
               sSpace = ""
            else
               sSpace = Space$(@rootNode.nodeLevel)
            end if
            sAccum = sAccum + sSpace + "</" + @rootNode.@ptrTag + ">" + $CRLF
         else
            sAccum = sAccum + "</" + @rootNode.@ptrTag + ">"
         end if
      end if
   end if

   Replace $cr with "" in sAccum
   Replace $lf with $crlf in sAccum
   Function = sAccum

End Function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzTreeToString
'  Formats an entire tree of XML data into its original string form.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzTreeToString Alias "xmlzTreeToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) Export as Dword

   g_xml_Return = xmlTreeToString(rootNode, iCRLF) + $nul
   Function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzBranchToString
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzBranchToString Alias "xmlzBranchToString" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iCRLF as Long ) Export as Dword

   g_xml_Return = xmlBranchToString(rootNode, iCRLF) + $nul
   Function = strptr(g_xml_Return)

End Function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlTreeToFile
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function xmlTreeToFile Alias "xmlTreeToFile" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal sFileName as String, _
        ByVal iOverwrite as Long ) export as Long

   Dim iFF as Long
   Dim sData as String

   '- Make sure the output file doesn't exist
   if dir$(sFileName) <> "" then
      if iOverwrite = %false then
         dd_xml_Error = "Output file exists: " + sFileName
         function = %false
         exit function
      else
         kill sFileName
         if dir$(sFileName) <> "" then
            dd_xml_Error = "Output file exists and cannot be deleted."
            function = %false
            exit function
         end if
      end if
   end if

   '- Open the output file
   errclear
   iFF = FreeFile
   Open sFileName for binary as #iFF
   if err then
      dd_xml_Error = "Unable to open output file '" + sFileName + "', err=" + format$(err)
      function = %false
      exit function
   end if

   sData = xmlTreeToString(ByVal rootNode, %false)
   put #iFF,, sData
   Close #iFF
   function = %true

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlNodeToFile
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlNodeToFile Alias "xmlNodeToFile" _
      ( rootNode as XML_NODE_TYPE PTR, _
        ByVal sFileName as String ) Export as Long

   Dim iFF as Long
   Dim sData as String

   '- Output file must not exist
   if dir$(sFileName) <> "" then
      dd_xml_Error = "File '" + sFileName + "' already exists."
      Function = %false
      exit function
   end if

   '- Turn the tree into a string
   sData = xmlTreeToString(rootNode, %false)

   '- Open the output file
   iFF = FreeFile
   Open sFileName for binary shared as #iFF
   if err then
      dd_xml_Error = "Unable to open output file '" + _
            sFileName + "' err=" + Format$(err)
      Function = %false
      exit function
   end if

   '- Write the data to the file
   Put #iFF,, sData
   Close #iFF

   Function = %true

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzNodeToFile
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzNodeToFile Alias "xmlzNodeToFile" _
      ( rootNode as XML_NODE_TYPE PTR, zFileName as Asciiz ) Export as Long

   Function = xmlNodeToFile(rootNode, (zFileName) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  internal_xmlReadAttributes                                     NOT EXPORTED
'  Takes everything after the tag specifier - <TAGNAME and
'  extracts the attributes of the tag. Sets iFoundEnd to
'  %True if this tag was self terminating ( <TAGNAME/> ) and
'  we won't find an end tag.
'  This is an internal function - not usually a need to call it directly
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub internal_xmlReadAttributes ( incoming as String, _
      xmlNode as XML_NODE_TYPE Ptr, ByVal termChar as String, iFoundEnd as Long )

   Dim iOffset as Long
   Dim bInQuotes as Long
   Dim bInValue as Long
   Dim sChar as String * 1
   Dim attribName as String
   Dim attribValue as String
   Dim quoteChar as String
   Dim lastChar as String * 1
   Dim curAttrib as XML_ATTRIB_TYPE PTR
   Dim nextAttrib as XML_ATTRIB_TYPE PTR
   Dim dummyAttrib as XML_ATTRIB_TYPE

   @xmlNode.ptrAttrib = 0

   iFoundEnd = %False
   iOffset = 1
   do until iOffset > len(incoming)

      sChar = mid$(incoming, iOffset, 1)

      if sChar = ">" then
         exit do
      end if

      '- We're reading the name not the value
      If bInValue = %False then
         if sChar = "=" then
            bInValue = %True
         elseif sChar = "/" then
            lastChar = "/"
         else
            attribName = attribName + sChar
         end if

      '- We're reading the value
      else

         '- Not yet in the quotes
         if bInQuotes = %False then

            '- Starting to read the value
            if (sChar = $DQ) or (sChar = "'") then
               quoteChar = sChar
               bInQuotes = %True
            end if
         else

            '- The end of the value
            if sChar = quoteChar then
               bInQuotes = %False
               bInValue = %False
               Gosub internal_xmlReadAttributes_AssignNode
               attribName = ""
               attribValue = ""

            '- We're reading the value
            else
               attribValue = attribValue + sChar
            end if

         end if
      end if
'      if (sChar <> " ") and (sChar <> $lf) and (sChar <> $cr) and (sChar <> $tab) then
'         lastChar = sChar
'      end if
      incr iOffset
   loop

   if lastChar = termChar then
      iFoundEnd = %True
   end if
   incoming = mid$(incoming, iOffset + 1)

exit sub

internal_xmlReadAttributes_AssignNode:
   nextAttrib = GetMem(len(@nextAttrib))
   if @xmlNode.ptrAttrib = 0 then
      @xmlNode.ptrAttrib = nextAttrib
   end if
   @nextAttrib.surroundChar = quoteChar

   Replace $lf with "" in attribName
   attribName = trim$(attribName)
   if len(attribName) then
      @nextAttrib.ptrName = GetMem(len(attribName) + 1)
      if @nextAttrib.ptrName then
         @nextAttrib.@ptrName = attribName
      end if
   end if

   if len(attribValue) then
      attribValue = xmlUnclean(attribValue)
      @nextAttrib.ptrValue = GetMem(len(attribValue) + 1)
      if @nextAttrib.ptrValue then
         @nextAttrib.@ptrValue = attribValue
      end if
   end if
   @nextAttrib.ptrNext = 0

   if curAttrib <> 0 then
      @curAttrib.ptrNext = nextAttrib
   end if
   curAttrib = nextAttrib
   Return

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  internal_xmlFreeAttribPointers                                 NOT EXPORTED
'  Recursively free the memory for the attribute linked list
'  Do not call this directly - it is called by xmlFreePointers
'  automatically.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub internal_xmlFreeAttribPointers(attribNode as XML_ATTRIB_TYPE PTR)

   Dim ptrAttrib as XML_ATTRIB_TYPE PTR

   if attribNode = 0 then exit sub

   if @attribNode.ptrNext <> 0 then
      ptrAttrib = @attribNode.ptrNext
      internal_xmlFreeAttribPointers ptrAttrib
   end if

   if @attribNode.ptrName <> 0 then
      #if %def(%DEBUG_FREE_PBCC)
      Stdout "FREE ATTRIB: " + @attribNode.@ptrName
      #endif
      FreeMem @attribNode.ptrName
      @attribNode.ptrName = 0
   end if

   if @attribNode.ptrValue <> 0 then
      FreeMem @attribNode.ptrValue
      @attribNode.ptrValue = 0
   end if

   FreeMem attribNode
   attribNode = 0

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFreePointers2
'  non-recursively clear all xml pointers
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub xmlFreePointers Alias "xmlFreePointers" _
      ( rootNode as XML_NODE_TYPE PTR ) Export

   Dim pNode as XML_NODE_TYPE PTR
   Dim pNext as XML_NODE_TYPE PTR

   pNode = rootNode
   do until pNode = 0

      '- Save the next node pointer
      pNext = @pNode.ptrNextAlloc

      '- Free mem for this node
      if @pNode.ptrTag then FreeMem @pNode.ptrTag
      if @pNode.ptrComment then FreeMem @pNode.ptrComment
      if @pNode.ptrAttrib then
         internal_xmlFreeAttribPointers @pNode.ptrAttrib
      end if
      FreeMem pNode

      pNode = pNext
   loop

End Sub

#if 0
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFreePointers
'  Recursively clear all xml pointers
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub xmlFreePointers Alias "xmlFreePointers" _
      ( rootNode as XML_NODE_TYPE Ptr ) Export

   if rootNode = 0 then
      exit sub
   end if

   '- Recursively free all children of this node.
   if @rootNode.ptrChild then
      xmlFreePointers @rootNode.ptrChild
   end if

   '- Free the next sibling
   if @rootNode.ptrNextSibling then
      xmlFreePointers @rootNode.ptrNextSibling
   end if

   '- Free mem for this node
   if @rootNode.ptrTag then FreeMem @rootNode.ptrTag
   if @rootNode.ptrComment then FreeMem @rootNode.ptrComment
   if @rootNode.ptrAttrib then
      internal_xmlFreeAttribPointers @rootNode.ptrAttrib
   end if

   FreeMem rootNode

End Sub
#endif

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFree
'  Alias for xmlFreePointers
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub xmlFree Alias "xmlFree" _
      ( rootNode as XML_NODE_TYPE Ptr ) Export

   xmlFreePointers rootNode

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFindNodeRecursive
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFindNodeRecursive Alias "xmlFindNodeRecursive" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iLevel as Long, _
        ByVal nodeTag as String, ByVal attribName as String, _
        ByVal attribValue as String ) Export as DWord

   Dim iFound as DWord

   if rootNode = 0 then
      Function = 0
      exit function
   end if

   if iLevel <> %XML_NODE_FIND_ANYWHERE then
      if (iLevel < @rootNode.nodeLevel) then
         Function = 0
         exit function
      end if
   end if

   iFound = 0

   '- Check this node
   if (iLevel = @rootNode.nodeLevel) or (iLevel = %XML_NODE_FIND_ANYWHERE) then
      if @rootNode.nodeType = %XML_NODE_TAG then
         if @rootNode.ptrTag <> 0 then
            if @rootNode.@ptrTag = nodeTag then
               if (attribName = "") then
                  iFound = rootnode
               elseif xmlAttribValue(rootNode, attribName) = attribValue then
                  iFound = rootnode
               end if
            end if
         end if
      end if
   end if

   '- Check the children
   if iFound = 0 then
      if (iLevel > @rootNode.nodeLevel) or (iLevel = %XML_NODE_FIND_ANYWHERE) then
         iFound = xmlFindNodeRecursive(@rootNode.ptrChild, iLevel, _
                                nodeTag, attribName, attribValue)
      end if
   end if

   '- Check siblings
   if iFound = 0 then
      iFound = xmlFindNodeRecursive(@rootNode.ptrNextSibling, iLevel, _
                           nodeTag, attribName, attribValue)
   end if

   Function = iFound

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFindNode
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFindNode Alias "xmlFindNode" _
      ( xRoot as XML_NODE_TYPE PTR, ByVal iLevel as Long, _
        ByVal nodeTag as String, ByVal attribName as String, _
        ByVal attribValue as String ) Export as DWord

   'Dim iFound as DWord
   Dim pFound as XML_NODE_TYPE PTR
   Dim pSearch as XML_NODE_TYPE PTR

   pFound = 0
   pSearch = xRoot

   do until pSearch = 0
      if (iLevel = %XML_NODE_FIND_ANYWHERE) or (iLevel = @pSearch.nodeLevel) then

         '- Only check this node if it is below the requested node
         '  or it is the requested node.
         '
         if @pSearch.nodeLevel => @xRoot.nodeLevel then
            if @pSearch.nodeType = %XML_NODE_TAG then
               if @pSearch.ptrTag <> 0 then
                  if @pSearch.@ptrTag = nodeTag then
                     if (attribName = "") then
                        pFound = pSearch
                        exit do
                     elseif xmlAttribValue(pSearch, attribName) = attribValue then
                        pFound = pSearch
                        exit do
                     end if
                  end if
               end if
            end if
         end if
      end if
      pSearch = @pSearch.ptrNextAlloc
   loop

   Function = pFound

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzFindNode
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzFindNode Alias "xmlzFindNode" _
      ( rootNode as XML_NODE_TYPE PTR, ByVal iLevel as Long, _
        nodeTag as Asciiz, attribName as Asciiz, _
        attribValue as Asciiz ) Export as DWord

   Function = xmlFindNode(rootNode, iLevel, (nodeTag), _
         (attribName), (attribValue) )

End Function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAddChild
'  Adds a node in as a child under the parent
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub xmlAddChild Alias "xmlAddChild" _
      ( parentNode as XML_NODE_TYPE PTR, childNode as XML_NODE_TYPE PTR ) _
      Export

   Dim lastNode as XML_NODE_TYPE PTR

   if parentNode = 0 then exit sub

   if @parentNode.ptrChild = 0 then
      @parentNode.ptrChild = childNode
      @childNode.ptrParent = parentNode
      @childNode.nodeLevel = @parentNode.nodeLevel + 1
   else
      lastNode = @parentNode.ptrChild
      Do until @lastNode.ptrNextSibling = 0
         lastNode = @lastNode.ptrNextSibling
      Loop
      @lastNode.ptrNextSibling = childNode
      @childNode.ptrPrevSibling = lastNode
      @childNode.ptrParent = @lastNode.ptrParent
      @childNode.nodeLevel = @parentNode.nodeLevel + 1
   end if

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAddSimpleChild
'     This function adds a tag node under a parent and then adds a
'     data node under that child. It returns a pointer to the tag node.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAddSimpleChild Alias "xmlAddSimpleChild" _
      ( parentNode as XML_NODE_TYPE PTR, ByVal simpleTag as String, _
        ByVal simpleValue as String ) Export as DWord

   Dim childNode as XML_NODE_TYPE PTR
   Dim childData as XML_NODE_TYPE PTR

   simpleTag = trim$(simpleTag)
   if simpleTag = "" then
      Goto xmlAddSimpleChild_Done
   end if

   if parentNode = 0 then
      Goto xmlAddSimpleChild_Done
   end if

   childNode = GetMem(len(@childNode))
   if childNode = 0 then
      Goto xmlAddSimpleChild_Done
   end if

   childData = GetMem(len(@childData))
   if childData = 0 then
      'xmlFreePointers childNode
      childNode = 0
      Goto xmlAddSimpleChild_Done
   end if

   @childNode.nodeType = %XML_NODE_TAG
   @childData.nodeType = %XML_NODE_DATA

   @childNode.nodeLevel = @parentNode.nodeLevel + 1
   @childData.nodeLevel = @childNode.nodeLevel + 2

   @childNode.selfTerminate = %False

   @childNode.ptrTag = GetAsciiz(simpleTag)
   @childData.ptrTag = GetAsciiz(simpleValue)

   xmlAddChild childNode, childData
   xmlAddChild parentNode, childNode

xmlAddSimpleChild_Done:
   Function = childNode

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzAddSimpleChild  (asciiz instead of strings)
'     This function adds a tag node under a parent and then adds a
'     data node under that child. It returns a pointer to the tag node.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzAddSimpleChild Alias "xmlzAddSimpleChild" _
      ( parentNode as XML_NODE_TYPE PTR, simpleTag as asciiz, _
        simpleValue as asciiz ) Export as DWord
   Function = xmlAddSimpleChild(parentNode, (simpleTag), (simpleValue) )
End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlGetTag
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlGetTag Alias "xmlGetTag" _
      (xmlNode as XML_NODE_TYPE PTR) Export as String

   if xmlNode = 0 then
      Function = ""
   elseif @xmlNode.ptrTag = 0 then
      Function = ""
   else
      Function = @xmlNode.@ptrTag
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzGetTag
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzGetTag Alias "xmlzGetTag" _
      (xmlNode as XML_NODE_TYPE PTR) Export as Dword

   g_xml_Return = xmlGetTag(xmlNode) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFindDataChild
'     Returns the first xml data node under the parent. 0 returned on error
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFindDataChild Alias "xmlFindDataChild" _
      ( parentNode as XML_NODE_TYPE PTR ) Export as DWord

   Dim childNode as XML_NODE_TYPE PTR
   Dim foundNode as XML_NODE_TYPE PTR

   if parentNode = 0 then
      Function = 0
      exit function
   end if

   if @parentNode.ptrChild = 0 then
      Function = 0
      exit function
   end if

   foundNode = 0
   childNode = @parentNode.ptrChild
   do until childNode = 0
      if @childNode.nodeType = %XML_NODE_DATA then
         foundNode = childNode
         exit do
      end if
      childNode = @childNode.ptrChild
   loop

   Function = foundNode

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlChildData
'     Returns the first data string child of the given node.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlChildData Alias "xmlChildData" _
      ( parentNode as XML_NODE_TYPE PTR ) Export as String

   Dim dataNode as XML_NODE_TYPE PTR
   dataNode = xmlFindDataChild(parentNode)
   if dataNode = 0 then
      Function = ""
   elseif @dataNode.ptrTag = 0 then
      Function = ""
   else
      Function = @dataNode.@ptrTag
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzChildData
'     Returns the first data string child of the given node.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzChildData Alias "xmlzChildData" _
      ( parentNode as XML_NODE_TYPE PTR ) Export as DWord

   g_xml_Return = xmlChildData(parentNode) + $nul
   function = strptr(g_xml_Return)

End Function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlExtractSimpleChildData
'     Finds the child whose tag is specifed. Under that child, it finds
'     the first data node and returns its data.
'
'     e.g. if node is <font> and the tree is ...
'     <font>
'     <font-size>12</font-size>
'     and you set sTag to be "font-size"
'     Returns 12 - the first data under the first child called <font-size>
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlExtractSimpleChildData Alias "xmlExtractSimpleChildData" _
      (parentNode as XML_NODE_TYPE PTR, ByVal sTag as String) Export as String

   Dim iLevel as Long
   Dim childNode as XML_NODE_TYPE PTR
   Dim dataNode as XML_NODE_TYPE PTR

   if parentNode = 0 then
      Function = ""
      exit function
   end if

   iLevel = @parentNode.nodeLevel + 1
   childNode = xmlFindNode(parentNode, iLevel, sTag, "", "")
   if childNode = 0 then
      function = ""
      exit function
   end if

   Function = xmlChildData(childNode)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzExtractSimpleChildData
'     Finds the child whose tag is specifed. Under that child, it finds
'     the first data node and returns its data.
'
'     e.g. if node is <font> and the tree is ...
'     <font>
'     <font-size>12</font-size>
'     and you set sTag to be "font-size"
'     Returns 12 - the first data under the first child called <font-size>
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzExtractSimpleChildData Alias "xmlzExtractSimpleChildData" _
      (parentNode as XML_NODE_TYPE PTR, zTag as Asciiz) Export as DWord

   g_xml_Return = xmlExtractSimpleChildData(parentNode, (zTag) ) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFindChild
'     Returns the first child with the specified tag.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFindChild Alias "xmlFindChild" _
      ( parentNode as XML_NODE_TYPE PTR, ByVal sTag as String) as Dword

   Dim foundNode as XML_NODE_TYPE PTR
   Dim childNode as XML_NODE_TYPE PTR

   if parentNode = 0 then
      Function = %False
      exit function
   end if

   sTag = trim$(sTag)
   foundNode = 0
   childNode = @parentNode.ptrChild
   Do until childNode = 0
      if xmlNodeType(childNode) = %XML_NODE_TAG then
         if zToString(@childNode.ptrTag) = sTag then
            foundNode = childNode
            exit do
         end if
      end if
      childNode = @childNode.ptrNextSibling
   Loop

   Function = foundNode

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzFindChild
'     Returns the first child with the specified tag.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzFindChild Alias "xmlzFindChild" _
      ( parentNode as XML_NODE_TYPE PTR, zTag as Asciiz) as Dword

   Function = xmlFindChild(parentNode, (zTag) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFindSpecificChild
'     Returns the first child with the specified tag and attribute
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFindSpecificChild Alias "xmlFindSpecificChild" _
      ( parentNode as XML_NODE_TYPE PTR, ByVal sTag as String, _
        ByVal sAttribName as String, ByVal sAttribValue as String) as Dword

   Dim foundNode as XML_NODE_TYPE PTR
   Dim childNode as XML_NODE_TYPE PTR

   if parentNode = 0 then
      Function = %False
      exit function
   end if

   sTag = trim$(sTag)
   foundNode = 0
   childNode = @parentNode.ptrChild
   Do until childNode = 0
      if @childNode.ptrTag <> 0 then
         if trim$(zToString(@childNode.ptrTag)) = sTag then
            if sAttribValue = xmlAttribValue(childNode, sAttribName) then
               foundNode = childNode
               exit do
            end if
         end if
      end if
      childNode = @childNode.ptrNextSibling
   Loop

   Function = foundNode

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzFindSpecificChild
'     Returns the first child with the specified tag and attribute
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzFindSpecificChild Alias "xmlzFindSpecificChild" _
      ( parentNode as XML_NODE_TYPE PTR, zTag as Asciiz, _
        zAttribName as String, zAttribValue as String) as Dword

   Function = xmlFindSpecificChild( parentNode, (zTag), _
         (zAttribName), (zAttribValue) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlIncreaseLevel
'     This routine increases the iLevel value of this node and all
'     children and siblings.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub xmlIncreaseLevel(ByVal pNode as XML_NODE_TYPE PTR, ByVal iHowMuch as Long)

   'Dim pNext as  XML_NODE_TYPE PTR

   if pNode = 0 then exit sub

   '- Recursively free all children of this node.
   if @pNode.ptrChild <> 0 then
      xmlIncreaseLevel ByVal @pNode.ptrChild, iHowMuch
   end if

   if @pNode.ptrNextSibling <> 0 then
      xmlIncreaseLevel ByVal @pNode.ptrNextSibling, iHowMuch
   end if

   '- Increase this node.
   @pNode.nodeLevel = @pNode.nodeLevel + iHowMuch

   '- Increase all siblings
   'pNext = @pNode.ptrNextSibling
   'Do until pNext = 0
   '
   '   '- Do the children of the sibling
   '   xmlIncreaseLevel ByVal @pNext.ptrChild, iHowMuch
   '
   '   '- Do the sibling
   '   @pNext.nodeLevel = @pNext.nodeLevel + iHowMuch
   '   pNext = @pNext.ptrNextSibling
   'Loop

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlFromList
'     Turns a LINKED_LIST_TYPE list into an XML string
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlFromList Alias "xmlFromList" _
      ( ByVal pList as LINKED_LIST_TYPE PTR, ByVal listName as String ) _
        export as String

   Dim sAccum as String
   Dim sName as String
   Dim sData as String
   Dim pNext as LINKED_LIST_TYPE PTR

   sAccum = "<list name=" + $dq + xmlClean(listName) + $dq + ">" + $crlf
   pNext = pList
   Do until pNext = 0
      sName = llName(pNext)
      sData = xmlClean(llData(pNext))
      sAccum = sAccum + "<listitem name=" + $dq + sName + $dq + ">" + _
               sData + "</listitem>" + $crlf
      pNext = @pNext.pNext
   Loop

   Function = sAccum + "</list>"

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzFromList
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzFromList Alias "xmlzFromList" _
      ( ByVal pList as LINKED_LIST_TYPE PTR, listName as Asciiz ) _
        export as Dword

   g_xml_Return = xmlFromList(pList, (listName) ) + $nul
   function = strptr(g_xml_Return)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlToList
'     Takes an xml string formatted linked list and returns a pointer
'     to a decoded linked list pointer.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlToList Alias "xmlToList" _
      ( pRoot as XML_NODE_TYPE PTR ) Export as Dword

   Dim iOk as Long
   Dim sName as String
   Dim sData as String
   Dim plReturn As LINKED_LIST_TYPE PTR
   Dim pFind as XML_NODE_TYPE PTR
   Dim pNext as XML_NODE_TYPE PTR

   plReturn = 0
   if pRoot = 0 then
      Goto xmlToList_Done
   end if

   pFind = xmlFindNode(pRoot, %XML_NODE_FIND_ANYWHERE, _
         "list", "", "")
   if pFind = 0 then
      dd_xml_Error = "List node not found in xml"
      Goto xmlToList_Done
   end if

   pNext = @pFind.ptrChild
   Do until pNext = 0
      if xmlGetTag(pNext) = "listitem" then
         sName = xmlAttribValue(pNext, "name")
         sData = xmlChildData(pNext)
         if plReturn = 0 then
            plReturn = llAdd(plReturn, sName, sData)
         else
            llUpdateByName plReturn, sName, sData
         end if
      end if
      pNext = @pNext.ptrNextSibling
   Loop

xmlToList_Done:

   Function = plReturn

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlNextSibling
'  Returns a pointer to the next sibling
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlNextSibling Alias "xmlNextSibling" _
      ( node as XML_NODE_TYPE PTR ) Export as Dword

   if node = 0 then
      Function = 0
   else
      Function = @node.ptrNextSibling
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlChild
'  Returns a pointer to the first child node
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlChild Alias "xmlChild" _
      ( node as XML_NODE_TYPE PTR ) Export as Dword

   if node = 0 then
      Function = 0
   else
      Function = @node.ptrChild
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlParent
'  Returns a pointer to the parent node
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlParent Alias "xmlParent" _
      ( node as XML_NODE_TYPE PTR ) Export as Dword

   if node = 0 then
      Function = 0
   else
      Function = @node.ptrParent
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlAttribToList
'  Converts all attributes for a node into a linked list and returns
'  a pointer to that list. The programmer is responsible for freeing
'  the memory used by the linked list with llFree when done.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlAttribToList ( node as XML_NODE_TYPE PTR ) export as Dword

   Dim pList as LINKED_LIST_TYPE PTR
   Dim pAttrib as XML_ATTRIB_TYPE PTR

   if node = 0 then
      function = 0
      exit function
   end if

   pList = 0

   pAttrib = @node.ptrAttrib
   Do until pAttrib = 0
      if pList = 0 then
         pList = llAdd(0, zToString(@pAttrib.ptrName), zToString(@pAttrib.ptrValue))
      else
         llUpdateByName pList, zToString(@pAttrib.ptrName), zToString(@pAttrib.ptrValue)
      end if
      pAttrib = @pAttrib.ptrNext
   Loop

   Function = pList

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlParse
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function xmlParse alias "xmlParse" (sXML as String) export as Dword

   Dim iOk as Long
   Dim iLevel as Long
   Dim iPos as Long
   Dim iLen as Long
   Dim iStatus as Long
   Dim iFoundEquals as Long
   Dim iSelfTerminate as Long
   Dim iCurrentLine as Long
   Dim iCurrentCol as Long

   Dim curChar as Byte
   Dim lastChar as Byte
   Dim lastlastChar as Byte
   Dim curDelim as Byte

   Dim sAccum as String
   Dim sName as String
   Dim sValue as String
   Dim sTag as String
   Dim criticalError as String

   Dim pXML as Byte Ptr

   Dim rootAttrib as XML_ATTRIB_TYPE PTR
   Dim curAttrib as XML_ATTRIB_TYPE PTR
   Dim lastAttrib as XML_ATTRIB_TYPE PTR

   Dim rootNode as XML_NODE_TYPE PTR
   Dim curNode as XML_NODE_TYPE PTR
   Dim lastNode as XML_NODE_TYPE PTR
   Dim lastAlloc as XML_NODE_TYPE PTR

   %IN_TAG_NAME = 1
   %IN_ATTRIB_BEFORE_NAME = 2
   %IN_ATTRIB_NAME = 3
   %IN_ATTRIB_BEFORE_VALUE = 4
   %IN_ATTRIB_VALUE = 5
   %IN_WHITE_SPACE = 6
   %IN_DATA = 7
   %IN_COMMENT = 8

   %XML_WHITESPACE_END = 33   ' 33 or greater is not white space, it's ascii
   %XML_TEXT_QUAL_1 = 34      ' "
   %XML_TEXT_QUAL_2 = 39      ' '
   %XML_BEGIN_TAG = 60        ' <
   %XML_END_TAG = 62          ' >
   %XML_EQUALS = 61           ' =
   %XML_SELF_TERMINATE = 47   ' /
   %XML_HYPHEN = 45           ' -

   iOk = %true

   pXML = strptr(sXML)
   iLen = len(sXML)
   if iLen < 1 then
      dd_xml_Error = "Empty xml send to xml parser"
      iOk = %false
      goto parseXML2_Done
   end if

   sAccum = ""
   iStatus = %IN_TAG_NAME
   lastChar = %XML_BEGIN_TAG

   '- Find our starting point
   iPos = 1
   do until @pXML[iPos-1] = %XML_BEGIN_TAG
      incr iPos
      if iPos > iLen then
         iOk = %false
         dd_xml_Error = "No starting tag found"
         goto parseXML2_Done
      end if
   loop

   iCurrentLine = 1

   '- Now do the parse
   do
      incr iPos
      if iPos > iLen then exit do
      lastChar = curChar
      curChar = @pXML[iPos-1]

      if curChar = 13 then
         incr iCurrentLine
         iCurrentCol = 1
      else
         incr iCurrentCol
      end if

      '- Process the current character
      Select Case iStatus

         '- Still getting the tag name.
         Case %IN_TAG_NAME

            '- If we hit white space then the tag must be done.
            if curChar < %XML_WHITESPACE_END then

               '- Error: no tag name specified
               if trim$(sTag) = "" then
                  iOk = %false
                  dd_xml_Error = "(" + format$(iCurrentLine) + ":" + Format$(iCurrentCol) + _
                        ") Empty tag name specified"
                  goto parseXML2_Done
               elseif left$(sTag, 3) = "!--" then
                  iStatus = %IN_COMMENT
               else
                  iStatus = %IN_ATTRIB_BEFORE_NAME
               end if

            else

               '- Special case - the tag is a comment
               if (curChar = %XML_HYPHEN) and (sTag = "!-") then
                  sTag = sTag + chr$(curChar)
                  iStatus = %IN_COMMENT

               else
                  Select Case curChar

                     '- Error invalid character in tag name
                     Case %XML_TEXT_QUAL_1, %XML_TEXT_QUAL_2, _
                          %XML_BEGIN_TAG, %XML_EQUALS
                        iOk = %false
                        dd_xml_Error = "(" + format$(iCurrentLine) + _
                              ":" + Format$(iCurrentCol) + ") " + _
                              "Invalid character found in tag name"
                        goto parseXML2_Done

                     '- The end of the tag
                     Case %XML_END_TAG
                        Gosub parseXML2_AddTag

                     '- Accumulate the name
                     Case else
                        sTag = sTag + chr$(curChar)

               End Select

               end if
            end if

         '- Starting a new attribute name.
         Case %IN_ATTRIB_BEFORE_NAME

            '- Start getting the name
            if curChar >= %XML_WHITESPACE_END then
               Select Case curChar

                  '- Error invalid character in tag
                  Case %XML_TEXT_QUAL_1, %XML_TEXT_QUAL_2, _
                       %XML_BEGIN_TAG, %XML_EQUALS
                     iOk = %false
                     dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Invalid character found in attribute name: " + _
                           chr$(curChar)
                     goto parseXML2_Done

                  '- The end of the tag
                  Case %XML_END_TAG
                     Gosub parseXML2_AddTag

                  Case %XML_SELF_TERMINATE
                     iSelfTerminate = %true

                  '- Accumulate the name
                  Case else
                     iSelfTerminate = %false
                     sName = chr$(curChar)
                     iStatus = %IN_ATTRIB_NAME

               End Select
            else
               '- do nothing. we ignore white space between tag
               '  attributes.
               '
            end if

         Case %IN_ATTRIB_NAME

            '- If we hit white space or =, then
            '  the attribute name must be done.
            '
            if (curChar < %XML_WHITESPACE_END) or _
               (curChar = %XML_EQUALS) then

               '- Error:   no attrib name specified
               '           this shouldn't ever happend
               '           because in order to start
               '           an attrib name, a value must
               '           have been found.
               '
               if trim$(sName) = "" then
                  iOk = %false
                  dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Empty attribute name found"
                  goto parseXML2_Done
               end if

               if curChar = %XML_EQUALS then
                  iFoundEquals = %true
               else
                  iFoundEquals = %false
               end if

               iStatus = %IN_ATTRIB_BEFORE_VALUE

            else
               Select Case curChar

                  '- Error invalid character in attribute name
                  Case %XML_TEXT_QUAL_1, %XML_TEXT_QUAL_2, _
                       %XML_BEGIN_TAG
                     iOk = %false
                     dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Invalid character found in attribute name (2): " + _
                           chr$(curChar)
                     goto parseXML2_Done

                  Case %XML_END_TAG
                     gosub parseXML2_AddTag

                  '- Accumulate the attribute name
                  Case else
                     sName = sName + chr$(curChar)

               End Select
            end if

         '- We have an attribute name and are waiting
         '  to find the " or ' to start the value.
         '
         Case %IN_ATTRIB_BEFORE_VALUE

            '- Ignore white space
            Select Case curChar
               Case %XML_EQUALS

                  '- Error - 2 equals signs found
                  if iFoundEquals then
                     iOk = %false
                     dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Double equals sign found in attribute"
                     goto parseXML2_Done
                  end if

                  iFoundEquals = %true

               Case %XML_TEXT_QUAL_1, %XML_TEXT_QUAL_2

                  '- Found an invalid character because
                  '  we don't have an equals sign yet.
                  '
                  if iFoundEquals = %false then
                     iOk = %false
                     dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "No equals sign found in attribute"
                     goto parseXML2_Done
                  end if

                  curDelim = curChar
                  sValue = ""
                  iStatus = %IN_ATTRIB_VALUE

               '- This is where they specify an attribute without
               '  a value. Though allowed in html, this is not
               '  allowed in xml.
               '
               Case %XML_END_TAG, %XML_SELF_TERMINATE
                  dd_xml_Error = "(" + format$(iCurrentLine) + _
                        ":" + Format$(iCurrentCol) + ") " + _
                        "No value specified for attribute '" + sName + "'"
                  iOk = %false
                  goto parseXML2_Done

               Case else

                  '- Error, invalid character found.
                  '  We ignore whitespace
                  '
                  if (curChar >= %XML_WHITESPACE_END) then
                     iOk = %false
                     dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Attribute invalid character before attribute value. Attrib=" + _
                           sName + " currentChar=" + format$(curChar)
                     goto parseXML2_Done
                  end if
            End Select

         Case %IN_ATTRIB_VALUE

            '- We're done with the value
            if curChar = curDelim then
               gosub parseXML2_AddAttrib

            '- Accumulate it
            else
               sValue = sValue + chr$(curChar)
            end if

         Case %IN_DATA
            Select Case curChar
               Case %XML_BEGIN_TAG
                  'if trim$(sAccum) = "" then
                  '   iStatus = %IN_TAG_NAME
                  'else
                     Gosub parseXML2_AddTag
                  'end if

               Case else
                  sAccum = sAccum + chr$(curChar)

            End Select

         Case %IN_COMMENT

            '- Check for the end of the comment
            if (lastChar = %XML_HYPHEN) and (lastlastChar = %XML_HYPHEN) and _
               (curChar = %XML_END_TAG) then
               gosub parseXML2_AddTag

            else
               sAccum = sAccum + chr$(curChar)

            end if

         '- INTERNAL ERROR - this shouldn't happen
         Case else
            iOk = %false
            dd_xml_Error = "(" + format$(iCurrentLine) + _
                           ":" + Format$(iCurrentCol) + ") " + _
                           "Internal error. Invalid status: " + Format$(iStatus)
            goto parseXML2_Done

      End Select

      '- Save the last two characters
      lastlastChar = lastChar
      lastChar = curChar

      if iLevel < 0 then
         dd_xml_Error = "(" + format$(iCurrentLine) + _
                        ":" + Format$(iCurrentCol) + ") " + _
                        criticalError
         iOk = %false
         goto parseXML2_Done
      end if
   loop

parseXML2_Done:
   sAccum = ""
   sName = ""
   sValue = ""
   sTag = ""

   if iOk = %false then
      if rootNode then
         xmlFreePointers rootNode
      end if
      function = 0
   else
      function = rootNode
   end if

EXIT FUNCTION

parseXML2_AddTag:

   '- Is this the end of a level?
   if left$(sTag, 1) = "/" then
      #if %def(%DEBUG_PBCC)
      stdout "ENDTAG:" + sTag
      #endif
      lastNode = curNode
      curNode = @curNode.ptrParent
      iLevel = iLevel - 1

      '- Force an error
      if curNode = 0 then
         criticalError = "Too many end tags found. Last was " + sTag
         iLevel = -1
      else
         if trim$(zToString(@curNode.ptrTag)) <> trim$(mid$(sTag, 2)) then
            iLevel = -1
            criticalError = "Nesting error. " + _
                  "Expecing end tag=" + trim$(zToString(@curNode.ptrTag)) + _
                  ", but found " + trim$(mid$(sTag, 2))
         end if
      end if
      iStatus = %IN_DATA
      iSelfTerminate = %false
      sAccum = ""
      sTag = ""
      sName = ""
      sValue = ""

   else

      '- Get memory for the new node
      if rootNode = 0 then
         rootNode = GetMem(len(@rootNode))
         lastAlloc = rootNode
         lastNode = 0
         curNode = rootNode
         @curNode.nodeLevel = 0
      else
         lastNode = curNode
         curNode = GetMem(len(@curNode))

         '- Last alloc is used for the sole purpose
         '  of freeing memory associated with nodes.
         '
         if lastAlloc then
            @lastAlloc.ptrNextAlloc = curNode
         end if
         lastAlloc = curNode
         @curNode.nodeLevel = iLevel
         @curNode.ptrNextAlloc = 0

         '- This node is a sibling of the last one
         if @lastNode.nodeLevel = iLevel then
            @lastNode.ptrNextSibling = curNode
            @curNode.ptrParent = @lastNode.ptrParent
            @curNode.ptrPrevSibling = lastNode

         '- This node is a child of the last one
         else
            if @lastNode.ptrChild = 0 then
               @lastNode.ptrChild = curNode
            end if
            @curNode.ptrParent = lastNode
            @curNode.ptrPrevSibling = 0
            @curNode.ptrNextSibling = 0
         end if
      end if

      '- Data node
      if sTag = "" then
         #if %def(%DEBUG_PBCC)
         stdout "DATA: " + format$(len(sAccum)) + " bytes"
         #endif
         @curNode.nodeType = %XML_NODE_DATA
         @curNode.ptrTag = GetAsciiz(xmlUnclean(sAccum))
         @curNode.selfTerminate = %true
         iStatus = %IN_TAG_NAME

      '- Comment node
      elseif left$(sTag, 3) = "!--" then
         #if %def(%DEBUG_PBCC)
         stdout "COMMENT: " +  format$(len(sAccum)) + " bytes"
         #endif
         @curNode.nodeType = %XML_NODE_COMMENT
         @curNode.ptrTag = GetAsciiz(sTag)
         sAccum = left$(sAccum, len(sAccum) - 2)
         @curNode.ptrComment = GetAsciiz(sAccum)
         @curNode.selfTerminate = %true
         iStatus = %IN_DATA

      '- XML node
      elseif left$(sTag, 1) = "?" then
         #if %def(%DEBUG_PBCC)
         stdout "XML: " + sTag
         #endif
         @curNode.nodeType = %XML_NODE_XML
         @curNode.ptrTag = GetAsciiz(sTag)
         @curNode.selfTerminate = %true
         @curNode.ptrAttrib = rootAttrib
         rootAttrib = 0
         iStatus = %IN_DATA

      '- Tag node
      else
         #if %def(%DEBUG_PBCC)
         stdout "TAG: " + sTag
         #endif
         @curNode.nodeType = %XML_NODE_TAG
         @curNode.ptrTag = GetAsciiz(sTag)
         @curNode.selfTerminate = iSelfTerminate
         @curNode.ptrAttrib = rootAttrib
         rootAttrib = 0
         iStatus = %IN_DATA
         #if %def(%DEBUG_PBCC)
         if iSelfTerminate then
            stdout "  self-terminated"
         end if

         #endif

      end if

      #if %def(%DEBUG_PBCC)
         stdout "  level:" + format$(iLevel);
      #endif

      if @curNode.selfTerminate = %false then
         iLevel = iLevel + 1
      end if
      #if %def(%DEBUG_PBCC)
      stdout ", " + Format$(iLevel)
      #endif
   end if

   sAccum = ""
   sTag = ""
   sName = ""
   sValue = ""
   iSelfTerminate = %false
   return

parseXML2_AddAttrib:
   #if %def(%DEBUG_PBCC)
   stdout "  attribute: " + xmlUnclean(sName) + "=" + xmlUnclean(sValue)
   #endif
   if (rootAttrib = 0) or (curAttrib = 0) then
      rootAttrib = getMem(len(@curAttrib))
      curAttrib = rootAttrib
   else
      lastAttrib = curAttrib
      curAttrib = getMem(len(@curAttrib))
      if lastAttrib then
         @lastAttrib.ptrNext = curAttrib
      end if
   end if

   @curAttrib.surroundChar = chr$(curDelim)
   @curAttrib.ptrName = GetAsciiz(xmlUnclean(sName))
   @curAttrib.ptrValue = GetAsciiz(xmlUnclean(sValue))

   iStatus = %IN_ATTRIB_BEFORE_NAME
   sName = ""
   sValue = ""
   sAccum = ""
   return

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzParse
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzParse Alias "xmlzParse" _
      ( incoming as Asciiz ) Export as DWord

   Function = xmlParse( (incoming) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlParseFileBeta
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlParseFile Alias "xmlParseFile" _
      ( fileName as String ) Export as Long

   Dim iFF as Long
   Dim sFile as String

   if dir$(fileName) = "" then
      dd_xml_error = "Cannot locate xml file " + fileName
      function = %false
      exit function
   end if

   iFF = FreeFile
   open fileName for binary shared as #iFF
   if err then
      dd_xml_error = "Error opening xml file " + fileName + _
            " error=" + Format$(err)
      function = %false
      exit function
   end if

   if lof(iFF) = 0 then
      dd_xml_error = "File " + fileName + " is empty."
      function = %false
      exit function
   end if

   sFile = Space$(lof(iFF))
   Get #iFF,, sFile
   Close #iFF

   function = xmlParse(sFile)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzParseFileBeta
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function xmlzParseFile Alias "xmlzParseFile" _
      ( zFileName as Asciiz ) Export as Dword

   Function = xmlParseFile( (zFileName) )

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlGetChildTagData
'
'  Returns the child data of a child tag. For example ...
'
'  <head><item1>This is a test</item1></head.
'  if pHead is a pointer to <head>, then
'  xmlGetChildTagData(pHead, "item1") returns "This is a test"
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function xmlGetChildTagData Alias "xmlGetChildTagData" _
      ( pRoot as XML_NODE_TYPE PTR, ByVal sTag as String ) export as String

   Dim pFind as XML_NODE_TYPE PTR

   if pRoot = 0 then
      function = ""
      exit function
   end if

   pFind = xmlFindNode(pRoot, %XML_ANYWHERE, sTag, "", "")
   if pFind = 0 then
      function = ""
   else
      function = xmlChildData(pFind)
   end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  xmlzGetChildTagData
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function xmlzGetChildTagData Alias "xmlzGetChildTagData" _
      ( pRoot as XML_NODE_TYPE PTR, zTag as Asciiz ) export as dword

   g_xml_Return = xmlGetChildTagData(pRoot, (zTag) ) + $nul
   function = strptr(g_xml_Return)

end function

#endif


